#!/usr/bin/python
# -*- coding: utf-8 -*-

from snapconnect import snap
from time import time
from time import sleep
import logging
import numpy
import ttyusb
import sys
import random
import string
import binascii
from snaplib import ScriptsManager
from snaplib import SnappyUploader
import os
import collections
try:
    from _version import __version__, __revision__
except ImportError:
    __version__ = "UNKNOWN"
    __revision__ = "UNKNOWN"


timeout_local_address = 1
timeout_set_pin_dir = 1
timeout_write_pin = 2

SPY_UPLOAD_TIMEOUT = -90


def randomword(length):
    return ''.join(random.choice(string.lowercase) for i in range(length))


class SnapEngine(snap.Snap):

    def __init__(
        self,
        port=None,
        expected_address=None,
        timeout=3,
        ports=None,
        level=logging.ERROR,
        addr='\x00\x00\x20',
        license_file='License.dat',
            nvparams_file='nvparams.dat'):
        super(SnapEngine, self).__init__(
            funcs={},
            license_file=license_file,
            nvparams_file=nvparams_file,
            addr=addr
            )
        self.opened_serial_ports = []
        # self.function_destination = None
        logging.basicConfig(level=level)  # messages to STDOUT
        self.retry = 15
        self.timeout = timeout
        # self.add_rpc_func('callback_function', self.callback_function)
        if port:
            print 'opening:', port
            self.open_serial(snap.SERIAL_TYPE_RS232, port)
        elif ports:
            for port in ports:
                print 'opening:', port
                self.open_serial(snap.SERIAL_TYPE_RS232, port)
        if expected_address:
            self.local_address = self.call(None, 'localAddr')
            if expected_address != self.local_address:
                raise Exception(
                    'ERROR: expected address is %s, but found %s on port %s',
                    (expected_address, self.local_address, port))
        # True WHILE a file upload is in progress
        self.progInProgress = False
        # Time to wait for a remote device to respond to upload requests
        self.wait_time = 100
        # Timer used to tell if commands are taking too long
        # (during the programming process)
        self.respTimer = 0
        # Flag that keeps the timer tick from rescheduling itself
        self.stopTimer = False
        # The following get filled in on-the-fly
        # Filled in by beginUpload()
        # The file to be transferred
        self.fileName = ''
        # SNAP Address of the remote device to update
        self.targetAddr = None
        self.errCode = None

    def upload_spy(
            self, mac, spy_file, unreset_delay=None, unreset_func=None,
            wait_time=100):
        self.wait_time = wait_time
        self.errCode = None
        val = None
        self.beginUpload(mac, spy_file)
        unreset_time = time() + unreset_delay if unreset_delay else None
        while self.uploader.progInProgress:
            self.eng.poll()
            if unreset_time:
                if time() >= unreset_time:
                    val = unreset_func() if isinstance(
                        unreset_func, collections.Callable
                        ) else unreset_func[0](*unreset_func[1:])
                    unreset_time = None
        if self.errCode != snap.SNAPPY_PROGRESS_COMPLETE:
            raise Exception(
                "spy_uploader_errCode = %s" %
                repr(self.errCode))
        return val

    def beginUpload(self, destAddr, filename):
        """
        Function that kicks off the script upload process
        Requires a SNAP Connect instance to be in place
        Requires a callback instance to be in place
        Parameters:
            destAddr  - the SNAP Address of the SNAP Node
                        to be uploaded with the script (SPY file)
            filename  - the exact file name (including path if needed)
                        of the SPY file to be uploaded
        """
        self.targetAddr = destAddr
        self.filename = filename
        self._startupload()

    def _startupload(self):
        """
        Called internally for every upload attempt.
        You should be calling beginUpload()
        """
        f = None
        uploadImage = None
        if not os.path.isfile(self.filename):
            raise ValueError('%s is not a file' % repr(self.filename))
        with open(self.filename, 'rb') as f:
            file_data = f.read()
        uploadImage = ScriptsManager.getSnappyStringFromExport(file_data)
        # Kickoff upload
        if uploadImage:
            self.respTimer = self.wait_time
            self.scheduler.schedule(1.0, self.timerTick)

            upload = self.snapConn.spy_upload_mgr.startUpload(
                self.targetAddr, uploadImage)
            upload.registerFinishedCallback(self.app_upload_hook)
            self.progInProgress = True
        return

    def app_upload_hook(self, snappy_upload_obj, result):
        if result == SnappyUploader.SNAPPY_PROGRESS_COMPLETE:
            logging.info('SPY upload success')
        elif result == snap.SNAPPY_PROGRESS_TIMEOUT:
            logging.error('SPY upload timeout error')
        else:
            logging.error('SPY upload error')

        self.completeProcess(result)

    def completeProcess(self, errCode):
        """
        Process has completed (either in failure or success)
        Reset for the next time
        """
        # Keep the next timer tick from rescheduling itself
        self.stopTimer = True
        self.progInProgress = False
        self.respTimer = 0  # Timer to tell if commands are taking too long
        self.progRetryNum = 0
        self.targetAddr = None  # SNAP Address of the remote device to update
        self.fileName = ''  # Filename too
        self.errCode = errCode
        self.ourCallback.finishedSpyUpgrade(errCode)

    def timerTick(self):
        """
        Function called by timing mechanism on a regular basis
        Keeps track of when retries are needed
        """
        if self.progInProgress is True:
            self.respTimer -= 1
            logging.info("timerTick %d", self.respTimer)
            if self.respTimer == 0:
                # Timeout uploading
                logging.error(
                    "Spy upload timeout to node: %s",
                    binascii.hexlify(self.targetAddr)
                    )
                self.completeProcess(SPY_UPLOAD_TIMEOUT)
                return False  # Do not reschedule the tick
        if self.stopTimer:  # The way to stop the timer if necessary
            self.stopTimer = False
            return False
        return True  # Reschedule yourself

    def call(self, dest, function_name, *args):
        cb_name = randomword(10)
        self.add_rpc_func(cb_name, self.callback_function)
        try:
            return self.call1(cb_name, dest, function_name, *args)
        finally:
            del self.local_dispatcher.funcs[cb_name]

    def call1(self, cb_name, dest, function_name, *args):
        # sys.stderr.write('call %s%s\n' % (function_name, repr(args)))
        # sys.stderr.write(str(self) + '\n')
        self.response_received = None
        self.function_destination = dest
        self.number_of_requests = 0
        while True:
            if self.number_of_requests > self.retry:
                raise Exception('ERROR: timeout calling %s' % function_name)
            self.number_of_requests += 1
            if self.number_of_requests > 1:
                # print 'request #%d calling %s' % (
                    # self.number_of_requests, function_name)
                if self.function_destination:
                    sys.stderr.write(
                        'request #%d: %s is calling: rpc(%s, %s, ...)\n' %
                        (
                            self.number_of_requests,
                            binascii.hexlify(self.local_addr()),
                            binascii.hexlify(self.function_destination),
                            function_name
                        )
                    )
            if self.function_destination:
                # sys.stderr.write(
                    # '%s calling: rpc %s\n' %
                    # (binascii.hexlify(self.local_addr()),
                        # binascii.hexlify(self.function_destination)))
                self.rpc(
                    self.function_destination,
                    'callback',
                    cb_name,
                    function_name, *args)
            else:
                # sys.stderr.write('mcast_rpc...\n')
                self.mcast_rpc(
                    1, 1, 'callback', cb_name,
                    function_name, *args)
            self.request_sent = time()
            function_response_timeout = self.timeout + self.request_sent
            while True:
                if time() >= function_response_timeout:
                    break
                self.poll()
                if self.response_received:
                    self.response_time = \
                        self.response_received - self.request_sent
                    # print 'response_time =', self.response_time
                    return self.return_value

    def print_info(self, myid, macinfo):
        print 'engine: ', myid
        print 'name: ', self.call(macinfo, 'loadNvParam', 8)
        local_address = self.call(macinfo, 'localAddr')
        local_address_str = (
            binascii.hexlify(local_address) if local_address else ''
            )
        print 'localAddr: ', local_address_str
        print 'channel: ', self.call(macinfo, 'loadNvParam', 4)
        nid = self.call(macinfo, 'loadNvParam', 3)
        nid_str = ('0x%.2x%.2x' % ((nid >> 8) % 256, nid % 256)) if nid else ''
        print 'networkId: ', nid_str
        fea = self.call(macinfo, 'loadNvParam', 11)
        fea_str = ('0x%.2x%.2x' % ((fea >> 8) % 256, fea % 256)) if fea else ''
        print 'feature: ', fea_str
        print

    def callback_function(self, return_value):
        # sys.stderr.write('callback_function return_value: %s\n' %
        # repr(return_value))
        # sys.stderr.write(str(self) + '\n')
        # sys.stderr.write('local_addr: %s\n' %
            # binascii.hexlify(self.local_addr()))
        if (
            self.function_destination is None or
                self.rpc_source_addr() == self.function_destination):
            self.response_received = time()
            self.return_value = return_value

    def read_voltage(self, mac, channel, poly, repeat=25):
        adcs = []
        while len(adcs) < repeat:
            adc = self.call(mac, 'readAdc', channel)
            # print adc
            adcs.append(float(adc))
        adc = sum(adcs) / len(adcs)
        # print 'adc =', adc
        return numpy.polyval(poly, adc)

    def open_serial(
        self,
        serial_type=snap.SERIAL_TYPE_RS232,
        port=None,
        vid=None,
        pid=None,
        serial_number=None,
        reconnect=False,
        dll_path=None,
        forward_groups=None,
        baudrate=None,
        io_loop=None,
            close_all=True):
        if port is None:
            if vid is None:
                raise ValueError("vid cannot be None if port is None")
            if pid is None:
                raise ValueError("pid cannot be None if port is None")
            port = ttyusb.find_tty_usb_sn(vid, pid, serial_number)
        if close_all:
            self.close_all_serial()
        if dll_path is not None:
            super(SnapEngine, self).open_serial(
                serial_type,
                port,
                reconnect=reconnect,
                dll_path=dll_path,
                forward_groups=forward_groups,
                baudrate=baudrate,
                io_loop=io_loop
                )
        else:
            super(SnapEngine, self).open_serial(
                serial_type,
                port,
                reconnect=reconnect,
                forward_groups=forward_groups,
                baudrate=baudrate,
                io_loop=io_loop
                )
        self.opened_serial_ports.append(port)

    def close_all_serial(self):
        for port in self.opened_serial_ports:
            self.close_serial(port)

    def open_usb_serial(
        self,
        usb_serial,
        reconnect=False,
        dll_path=None,
        forward_groups=None,
        baudrate=None,
        io_loop=None,
            close_all=True):
        self.open_serial(
            vid=usb_serial['vid'],
            pid=usb_serial['pid'],
            serial_number=usb_serial['sn'],
            reconnect=reconnect,
            dll_path=dll_path,
            forward_groups=forward_groups,
            baudrate=baudrate,
            io_loop=io_loop,
            close_all=close_all,
        )

    def read_snappy_os_version(self, mac):
        version_major = self.call(mac, 'getInfo', 5)
        version_minor = self.call(mac, 'getInfo', 6)
        version_build = self.call(mac, 'getInfo', 7)
        return (version_major, version_minor, version_build)

    def enable_crc(self, mac, try_enabled_first=False, enable=True):
        mac_str = 'Unknown' if mac is None else binascii.hexlify(mac)
        retry_save = self.retry
        self.retry = 1
        comm_retry = 0
        while True:
            comm_retry += 1
            if comm_retry > retry_save:
                self.retry = retry_save
                raise Exception(
                    'comm failed with node: %s' % mac_str
                    )
            if try_enabled_first:
                try:
                    self.save_nv_param(11, 0x100)
                    print 'loadNvParam(11)...',
                    nv11 = self.call(mac, 'loadNvParam', 11)
                    print '0x%x' % nv11
                    if enable:
                        break
                    print 'save'
                    self.call(
                        mac,
                        'saveNvParam',
                        11,
                        0xfeff & nv11
                        )
                    print 'reboot'
                    self.call(
                        mac,
                        'reboot',
                        )
                    sleep(2)
                    self.save_nv_param(11, 0)
                    break
                except:
                    print(
                        'with CRC enabled, comm failed with node %s' %
                        mac_str
                        )
                try:
                    self.save_nv_param(11, 0)
                    print 'loadNvParam(11)...',
                    nv11 = self.call(mac, 'loadNvParam', 11)
                    print '0x%x' % nv11
                    if not enable:
                        break
                    print 'save'
                    self.call(
                        mac,
                        'saveNvParam',
                        11,
                        0x100 | nv11
                        )
                    print 'reboot'
                    self.call(
                        mac,
                        'reboot',
                        )
                    sleep(2)
                    self.save_nv_param(11, 0x100)
                    break
                except:
                    print(
                        'with CRC disabled, comm failed with node %s' %
                        mac_str
                        )
            else:
                try:
                    self.save_nv_param(11, 0)
                    print 'loadNvParam(11)...',
                    nv11 = self.call(mac, 'loadNvParam', 11)
                    print '0x%x' % nv11
                    if not enable:
                        break
                    print 'save'
                    self.call(
                        mac,
                        'saveNvParam',
                        11,
                        0x100 | nv11
                        )
                    print 'reboot'
                    self.call(
                        mac,
                        'reboot',
                        )
                    sleep(2)
                    self.save_nv_param(11, 0x100)
                    break
                except:
                    print(
                        'with CRC disabled, comm failed with node %s' %
                        mac_str
                        )
                try:
                    self.save_nv_param(11, 0x100)
                    print 'loadNvParam(11)...',
                    nv11 = self.call(mac, 'loadNvParam', 11)
                    print '0x%x' % nv11
                    if enable:
                        break
                    print 'save'
                    self.call(
                        mac,
                        'saveNvParam',
                        11,
                        0xfeff & nv11
                        )
                    print 'reboot'
                    self.call(
                        mac,
                        'reboot',
                        )
                    sleep(2)
                    self.save_nv_param(11, 0)
                    break
                except:
                    print(
                        'with CRC enabled, comm failed with node %s' %
                        mac_str
                        )
        self.retry = retry_save

    def rpcpoll(self, dest, function_name, *args):
        self.rpc(dest, function_name, *args)
        self.poll()


def _main():
    logging.basicConfig(level=logging.INFO)
    logging.info('snapengine version: %s', __version__)
    logging.info('snapengine revision: %s', __revision__)


if __name__ == "__main__":
    _main()
